@@ -34,10 +34,12 @@
 **
 ****************************************************************************/
 
+#include "qquickcontrol_p_p.h"
 #include "qquickoverlay_p.h"
 #include "qquickoverlay_p_p.h"
 #include "qquickpopupitem_p_p.h"
 #include "qquickpopup_p_p.h"
+#include "qquickdrawer_p.h"
 #include "qquickdrawer_p_p.h"
 #include "qquickapplicationwindow_p.h"
 #include <QtQml/qqmlinfo.h>
@@ -47,102 +49,25 @@
 
 QT_BEGIN_NAMESPACE
 
-void QQuickOverlayPrivate::popupAboutToShow()
-{
-    Q_Q(QQuickOverlay);
-    QQuickPopup *popup = qobject_cast<QQuickPopup *>(q->sender());
-    if (!popup || !popup->dim())
-        return;
+/*!
+    \qmltype Overlay
+    \inherits Item
+//!     \instantiates QQuickOverlay
+    \inqmlmodule QtQuick.Controls
+    \since 5.10
+    \brief A window overlay for popups.
 
-    // use QQmlProperty instead of QQuickItem::setOpacity() to trigger QML Behaviors
-    QQuickPopupPrivate *p = QQuickPopupPrivate::get(popup);
-    if (p->dimmer)
-        QQmlProperty::write(p->dimmer, QStringLiteral("opacity"), 1.0);
-}
+    Overlay provides a layer for popups, ensuring that popups are displayed above
+    other content and that the background is dimmed when a \l {Popup::}{modal} or
+    \l {Popup::dim}{dimmed} popup is visible.
 
-void QQuickOverlayPrivate::popupAboutToHide()
-{
-    Q_Q(QQuickOverlay);
-    QQuickPopup *popup = qobject_cast<QQuickPopup *>(q->sender());
-    if (!popup || !popup->dim())
-        return;
+    The overlay is an ordinary Item that covers the entire window. It can be used
+    as a visual parent to position a popup in scene coordinates.
 
-    // use QQmlProperty instead of QQuickItem::setOpacity() to trigger QML Behaviors
-    QQuickPopupPrivate *p = QQuickPopupPrivate::get(popup);
-    if (p->dimmer)
-        QQmlProperty::write(p->dimmer, QStringLiteral("opacity"), 0.0);
-}
+    \include qquickoverlay-popup-parent.qdocinc
 
-static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQuickItem *parent)
-{
-    QQuickItem *item = nullptr;
-    if (component) {
-        QQmlContext *creationContext = component->creationContext();
-        if (!creationContext)
-            creationContext = qmlContext(popup);
-        QQmlContext *context = new QQmlContext(creationContext, popup);
-        context->setContextObject(popup);
-        item = qobject_cast<QQuickItem*>(component->beginCreate(context));
-    }
-
-    // when there is no overlay component available (with plain QQuickWindow),
-    // use a plain QQuickItem as a fallback to block hover events
-    if (!item && popup->isModal())
-        item = new QQuickItem;
-
-    if (item) {
-        item->setOpacity(popup->isVisible() ? 1.0 : 0.0);
-        item->setParentItem(parent);
-        item->stackBefore(popup->popupItem());
-        item->setZ(popup->z());
-        if (popup->isModal()) {
-            item->setAcceptedMouseButtons(Qt::AllButtons);
-#if QT_CONFIG(cursor)
-            item->setCursor(Qt::ArrowCursor);
-#endif
-#if QT_CONFIG(quicktemplates2_hover)
-            // TODO: switch to QStyleHints::useHoverEffects in Qt 5.8
-            item->setAcceptHoverEvents(true);
-            // item->setAcceptHoverEvents(QGuiApplication::styleHints()->useHoverEffects());
-            // connect(QGuiApplication::styleHints(), &QStyleHints::useHoverEffectsChanged, item, &QQuickItem::setAcceptHoverEvents);
-#endif
-        }
-        if (component)
-            component->completeCreate();
-    }
-    return item;
-}
-
-void QQuickOverlayPrivate::createOverlay(QQuickPopup *popup)
-{
-    Q_Q(QQuickOverlay);
-    QQuickPopupPrivate *p = QQuickPopupPrivate::get(popup);
-    if (!p->dimmer)
-        p->dimmer = createDimmer(popup->isModal() ? modal : modeless, popup, q);
-    p->resizeOverlay();
-}
-
-void QQuickOverlayPrivate::destroyOverlay(QQuickPopup *popup)
-{
-    QQuickPopupPrivate *p = QQuickPopupPrivate::get(popup);
-    if (p->dimmer) {
-        p->dimmer->setParentItem(nullptr);
-        p->dimmer->deleteLater();
-        p->dimmer = nullptr;
-    }
-}
-
-void QQuickOverlayPrivate::toggleOverlay()
-{
-    Q_Q(QQuickOverlay);
-    QQuickPopup *popup = qobject_cast<QQuickPopup *>(q->sender());
-    if (!popup)
-        return;
-
-    destroyOverlay(popup);
-    if (popup->dim())
-        createOverlay(popup);
-}
+    \sa ApplicationWindow
+*/
 
 QVector<QQuickPopup *> QQuickOverlayPrivate::stackingOrderPopups() const
 {
@@ -174,12 +99,6 @@ void QQuickOverlayPrivate::itemGeometryChanged(QQuickItem *, QQuickGeometryChang
     updateGeometry();
 }
 
-QQuickOverlayPrivate::QQuickOverlayPrivate()
-    : modal(nullptr),
-      modeless(nullptr)
-{
-}
-
 bool QQuickOverlayPrivate::startDrag(QEvent *event, const QPointF &pos)
 {
     Q_Q(QQuickOverlay);
@@ -209,11 +128,6 @@ bool QQuickOverlayPrivate::startDrag(QEvent *event, const QPointF &pos)
     return false;
 }
 
-static bool isTouchEvent(QEvent *event)
-{
-    return event->type() == QEvent::TouchBegin || event->type() == QEvent::TouchUpdate || event->type() == QEvent::TouchEnd;
-}
-
 bool QQuickOverlayPrivate::handlePress(QQuickItem *source, QEvent *event, QQuickPopup *target)
 {
     if (target) {
@@ -222,7 +136,18 @@ bool QQuickOverlayPrivate::handlePress(QQuickItem *source, QEvent *event, QQuick
             return true;
         }
         return false;
-    } else if (!mouseGrabberPopup || isTouchEvent(event)) {
+    }
+
+    switch (event->type()) {
+    default: {
+        if (mouseGrabberPopup)
+            break;
+#if QT_CONFIG(quicktemplates2_multitouch)
+        Q_FALLTHROUGH();
+    case QEvent::TouchBegin:
+    case QEvent::TouchUpdate:
+    case QEvent::TouchEnd:
+#endif
         // allow non-modal popups to close themselves,
         // and non-dimming modal popups to block the event
         const auto popups = stackingOrderPopups();
@@ -232,6 +157,8 @@ bool QQuickOverlayPrivate::handlePress(QQuickItem *source, QEvent *event, QQuick
                 return true;
             }
         }
+        break;
+    }
     }
 
     event->ignore();
@@ -330,7 +257,7 @@ void QQuickOverlayPrivate::removePopup(QQuickPopup *popup)
 {
     Q_Q(QQuickOverlay);
     allPopups.removeOne(popup);
-    if (allDrawers.removeOne(static_cast<QQuickDrawer *>(popup)))
+    if (allDrawers.removeOne(qobject_cast<QQuickDrawer *>(popup)))
         q->setVisible(!allDrawers.isEmpty() || !q->childItems().isEmpty());
 }
 
@@ -384,6 +311,9 @@ QQuickOverlay::QQuickOverlay(QQuickItem *parent)
     Q_D(QQuickOverlay);
     setZ(1000001); // DefaultWindowDecoration+1
     setAcceptedMouseButtons(Qt::AllButtons);
+#if QT_CONFIG(quicktemplates2_multitouch)
+    setAcceptTouchEvents(true);
+#endif
     setFiltersChildMouseEvents(true);
     setVisible(false);
 
@@ -416,7 +346,6 @@ void QQuickOverlay::setModal(QQmlComponent *modal)
     if (d->modal == modal)
         return;
 
-    delete d->modal;
     d->modal = modal;
     emit modalChanged();
 }
@@ -433,7 +362,6 @@ void QQuickOverlay::setModeless(QQmlComponent *modeless)
     if (d->modeless == modeless)
         return;
 
-    delete d->modeless;
     d->modeless = modeless;
     emit modelessChanged();
 }
@@ -453,7 +381,7 @@ QQuickOverlay *QQuickOverlay::overlay(QQuickWindow *window)
         QQuickItem *content = window->contentItem();
         // Do not re-create the overlay if the window is being destroyed
         // and thus, its content item no longer has a window associated.
-        if (content->window()) {
+        if (content && content->window()) {
             overlay = new QQuickOverlay(window->contentItem());
             window->setProperty(name, QVariant::fromValue(overlay));
         }
@@ -461,37 +389,18 @@ QQuickOverlay *QQuickOverlay::overlay(QQuickWindow *window)
     return overlay;
 }
 
+QQuickOverlayAttached *QQuickOverlay::qmlAttachedProperties(QObject *object)
+{
+    return new QQuickOverlayAttached(object);
+}
+
 void QQuickOverlay::itemChange(ItemChange change, const ItemChangeData &data)
 {
     Q_D(QQuickOverlay);
     QQuickItem::itemChange(change, data);
 
-    QQuickPopup *popup = nullptr;
-    if (change == ItemChildAddedChange || change == ItemChildRemovedChange) {
-        popup = qobject_cast<QQuickPopup *>(data.item->parent());
+    if (change == ItemChildAddedChange || change == ItemChildRemovedChange)
         setVisible(!d->allDrawers.isEmpty() || !childItems().isEmpty());
-    }
-    if (!popup)
-        return;
-
-    if (change == ItemChildAddedChange) {
-        if (popup->dim())
-            d->createOverlay(popup);
-        QObjectPrivate::connect(popup, &QQuickPopup::dimChanged, d, &QQuickOverlayPrivate::toggleOverlay);
-        QObjectPrivate::connect(popup, &QQuickPopup::modalChanged, d, &QQuickOverlayPrivate::toggleOverlay);
-        if (!qobject_cast<QQuickDrawer *>(popup)) {
-            QObjectPrivate::connect(popup, &QQuickPopup::aboutToShow, d, &QQuickOverlayPrivate::popupAboutToShow);
-            QObjectPrivate::connect(popup, &QQuickPopup::aboutToHide, d, &QQuickOverlayPrivate::popupAboutToHide);
-        }
-    } else if (change == ItemChildRemovedChange) {
-        d->destroyOverlay(popup);
-        QObjectPrivate::disconnect(popup, &QQuickPopup::dimChanged, d, &QQuickOverlayPrivate::toggleOverlay);
-        QObjectPrivate::disconnect(popup, &QQuickPopup::modalChanged, d, &QQuickOverlayPrivate::toggleOverlay);
-        if (!qobject_cast<QQuickDrawer *>(popup)) {
-            QObjectPrivate::disconnect(popup, &QQuickPopup::aboutToShow, d, &QQuickOverlayPrivate::popupAboutToShow);
-            QObjectPrivate::disconnect(popup, &QQuickPopup::aboutToHide, d, &QQuickOverlayPrivate::popupAboutToHide);
-        }
-    }
 }
 
 void QQuickOverlay::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
@@ -654,4 +563,150 @@ bool QQuickOverlay::eventFilter(QObject *object, QEvent *event)
     return false;
 }
 
+class QQuickOverlayAttachedPrivate : public QObjectPrivate
+{
+    Q_DECLARE_PUBLIC(QQuickOverlayAttached)
+
+public:
+    void setWindow(QQuickWindow *newWindow);
+
+    QQuickWindow *window = nullptr;
+    QQmlComponent *modal = nullptr;
+    QQmlComponent *modeless = nullptr;
+};
+
+void QQuickOverlayAttachedPrivate::setWindow(QQuickWindow *newWindow)
+{
+    Q_Q(QQuickOverlayAttached);
+    if (window == newWindow)
+        return;
+
+    if (QQuickOverlay *oldOverlay = QQuickOverlay::overlay(window)) {
+        QObject::disconnect(oldOverlay, &QQuickOverlay::pressed, q, &QQuickOverlayAttached::pressed);
+        QObject::disconnect(oldOverlay, &QQuickOverlay::released, q, &QQuickOverlayAttached::released);
+    }
+
+    if (QQuickOverlay *newOverlay = QQuickOverlay::overlay(newWindow)) {
+        QObject::connect(newOverlay, &QQuickOverlay::pressed, q, &QQuickOverlayAttached::pressed);
+        QObject::connect(newOverlay, &QQuickOverlay::released, q, &QQuickOverlayAttached::released);
+    }
+
+    window = newWindow;
+    emit q->overlayChanged();
+}
+
+/*!
+    \qmlattachedsignal QtQuick.Controls::Overlay::pressed()
+
+    This attached signal is emitted when the overlay is pressed by the user while
+    a popup is visible.
+
+    The signal can be attached to any item, popup, or window. When attached to an
+    item or a popup, the signal is only emitted if the item or popup is in a window.
+*/
+
+/*!
+    \qmlattachedsignal QtQuick.Controls::Overlay::released()
+
+    This attached signal is emitted when the overlay is released by the user while
+    a popup is visible.
+
+    The signal can be attached to any item, popup, or window. When attached to an
+    item or a popup, the signal is only emitted if the item or popup is in a window.
+*/
+
+QQuickOverlayAttached::QQuickOverlayAttached(QObject *parent)
+    : QObject(*(new QQuickOverlayAttachedPrivate), parent)
+{
+    Q_D(QQuickOverlayAttached);
+    if (QQuickItem *item = qobject_cast<QQuickItem *>(parent)) {
+        d->setWindow(item->window());
+        QObjectPrivate::connect(item, &QQuickItem::windowChanged, d, &QQuickOverlayAttachedPrivate::setWindow);
+    } else if (QQuickPopup *popup = qobject_cast<QQuickPopup *>(parent)) {
+        d->setWindow(popup->window());
+        QObjectPrivate::connect(popup, &QQuickPopup::windowChanged, d, &QQuickOverlayAttachedPrivate::setWindow);
+    } else {
+        d->setWindow(qobject_cast<QQuickWindow *>(parent));
+    }
+}
+
+/*!
+    \qmlattachedproperty Overlay QtQuick.Controls::Overlay::overlay
+    \readonly
+
+    This attached property holds the window overlay item.
+
+    The property can be attached to any item, popup, or window. When attached to an
+    item or a popup, the value is \c null if the item or popup is not in a window.
+*/
+QQuickOverlay *QQuickOverlayAttached::overlay() const
+{
+    Q_D(const QQuickOverlayAttached);
+    return QQuickOverlay::overlay(d->window);
+}
+
+/*!
+    \qmlattachedproperty Component QtQuick.Controls::Overlay::modal
+
+    This attached property holds a component to use as a visual item that implements
+    background dimming for modal popups. It is created for and stacked below visible
+    modal popups.
+
+    The property can be attached to any popup.
+
+    For example, to change the color of the background dimming for a modal
+    popup, the following code can be used:
+
+    \snippet qtquickcontrols2-overlay-modal.qml 1
+
+    \sa Popup::modal
+*/
+QQmlComponent *QQuickOverlayAttached::modal() const
+{
+    Q_D(const QQuickOverlayAttached);
+    return d->modal;
+}
+
+void QQuickOverlayAttached::setModal(QQmlComponent *modal)
+{
+    Q_D(QQuickOverlayAttached);
+    if (d->modal == modal)
+        return;
+
+    d->modal = modal;
+    emit modalChanged();
+}
+
+/*!
+    \qmlattachedproperty Component QtQuick.Controls::Overlay::modeless
+
+    This attached property holds a component to use as a visual item that implements
+    background dimming for modeless popups. It is created for and stacked below visible
+    dimming popups.
+
+    The property can be attached to any popup.
+
+    For example, to change the color of the background dimming for a modeless
+    popup, the following code can be used:
+
+    \snippet qtquickcontrols2-overlay-modeless.qml 1
+
+    \sa Popup::dim
+*/
+QQmlComponent *QQuickOverlayAttached::modeless() const
+{
+    Q_D(const QQuickOverlayAttached);
+    return d->modeless;
+}
+
+void QQuickOverlayAttached::setModeless(QQmlComponent *modeless)
+{
+    Q_D(QQuickOverlayAttached);
+    if (d->modeless == modeless)
+        return;
+
+    d->modeless = modeless;
+    emit modelessChanged();
+}
+
 QT_END_NAMESPACE
