@@ -38,10 +38,13 @@
 #include <QtTest/qsignalspy.h>
 #include "../shared/util.h"
 #include "../shared/visualtestutil.h"
+#include "../shared/qtest_quickcontrols.h"
 
 #include <QtGui/qpa/qwindowsysteminterface.h>
+#include <QtQuick/qquickview.h>
 #include <QtQuickTemplates2/private/qquickapplicationwindow_p.h>
 #include <QtQuickTemplates2/private/qquickcombobox_p.h>
+#include <QtQuickTemplates2/private/qquickdialog_p.h>
 #include <QtQuickTemplates2/private/qquickoverlay_p.h>
 #include <QtQuickTemplates2/private/qquickpopup_p.h>
 #include <QtQuickTemplates2/private/qquickbutton_p.h>
@@ -69,6 +72,7 @@ private slots:
     void activeFocusOnClose1();
     void activeFocusOnClose2();
     void activeFocusOnClose3();
+    void activeFocusOnClosingSeveralPopups();
     void hover_data();
     void hover();
     void wheel_data();
@@ -79,8 +83,20 @@ private slots:
     void cursorShape();
     void componentComplete();
     void closeOnEscapeWithNestedPopups();
+    void closeOnEscapeWithVisiblePopup();
+    void enabled();
     void orientation_data();
     void orientation();
+    void qquickview();
+    void disabledPalette();
+    void disabledParentPalette();
+    void countChanged();
+    void toolTipCrashOnClose();
+    void setOverlayParentToNull();
+    void tabFence();
+    void invisibleToolTipOpen();
+    void centerInOverlayWithinStackViewItem();
+    void destroyDuringExitTransition();
 };
 
 void tst_QQuickPopup::initTestCase()
@@ -100,6 +116,7 @@ void tst_QQuickPopup::visible()
 {
     QFETCH(QString, source);
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -118,7 +135,7 @@ void tst_QQuickPopup::visible()
     QVERIFY(overlay->childItems().contains(popupItem));
 
     popup->close();
-    QVERIFY(!popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
     QVERIFY(!overlay->childItems().contains(popupItem));
 
     popup->setVisible(true);
@@ -126,13 +143,14 @@ void tst_QQuickPopup::visible()
     QVERIFY(overlay->childItems().contains(popupItem));
 
     popup->setVisible(false);
-    QVERIFY(!popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
     QVERIFY(!overlay->childItems().contains(popupItem));
 }
 
 void tst_QQuickPopup::state()
 {
     QQuickApplicationHelper helper(this, "applicationwindow.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -163,7 +181,7 @@ void tst_QQuickPopup::state()
     QCOMPARE(closedSpy.count(), 0);
 
     popup->close();
-    QCOMPARE(visibleChangedSpy.count(), 2);
+    QTRY_COMPARE(visibleChangedSpy.count(), 2);
     QCOMPARE(aboutToShowSpy.count(), 1);
     QCOMPARE(aboutToHideSpy.count(), 1);
     QCOMPARE(openedSpy.count(), 1);
@@ -194,6 +212,7 @@ void tst_QQuickPopup::overlay()
     QFETCH(bool, dim);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -217,6 +236,15 @@ void tst_QQuickPopup::overlay()
     QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
     QVERIFY(popup);
 
+    QQuickOverlayAttached *overlayAttached = qobject_cast<QQuickOverlayAttached *>(qmlAttachedPropertiesObject<QQuickOverlay>(popup));
+    QVERIFY(overlayAttached);
+    QCOMPARE(overlayAttached->overlay(), overlay);
+
+    QSignalSpy overlayAttachedPressedSignal(overlayAttached, SIGNAL(pressed()));
+    QSignalSpy overlayAttachedReleasedSignal(overlayAttached, SIGNAL(released()));
+    QVERIFY(overlayAttachedPressedSignal.isValid());
+    QVERIFY(overlayAttachedReleasedSignal.isValid());
+
     QQuickButton *button = window->property("button").value<QQuickButton*>();
     QVERIFY(button);
 
@@ -226,18 +254,22 @@ void tst_QQuickPopup::overlay()
     popup->open();
     QVERIFY(popup->isVisible());
     QVERIFY(overlay->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), ++overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), overlayReleaseCount);
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
+
+    QTRY_VERIFY(!popup->isVisible());
+    QVERIFY(!overlay->isVisible());
 
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), overlayReleaseCount); // no modal-popups open
-
-    popup->close();
-    QVERIFY(!popup->isVisible());
-    QVERIFY(!overlay->isVisible());
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
 
     popup->setDim(dim);
     popup->setModal(modal);
@@ -247,17 +279,22 @@ void tst_QQuickPopup::overlay()
     popup->open();
     QVERIFY(popup->isVisible());
     QVERIFY(overlay->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), ++overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), overlayReleaseCount);
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
 
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), ++overlayReleaseCount);
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
 
-    QVERIFY(!popup->isVisible());
-    QCOMPARE(overlay->isVisible(), popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
+    QVERIFY(!overlay->isVisible());
 
     // touch
     popup->open();
@@ -280,13 +317,17 @@ void tst_QQuickPopup::overlay()
     QTest::touchEvent(window, device.data()).press(0, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), ++overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), overlayReleaseCount);
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
 
     QTest::touchEvent(window, device.data()).release(0, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), overlayPressCount);
     QCOMPARE(overlayReleasedSignal.count(), ++overlayReleaseCount);
+    QCOMPARE(overlayAttachedPressedSignal.count(), overlayPressCount);
+    QCOMPARE(overlayAttachedReleasedSignal.count(), overlayReleaseCount);
 
-    QVERIFY(!popup->isVisible());
-    QCOMPARE(overlay->isVisible(), popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
+    QVERIFY(!overlay->isVisible());
 
     // multi-touch
     popup->open();
@@ -309,7 +350,7 @@ void tst_QQuickPopup::overlay()
     QCOMPARE(overlayReleasedSignal.count(), overlayReleaseCount);
 
     QTest::touchEvent(window, device.data()).release(0, button->mapToScene(QPointF(1, 1)).toPoint()).stationary(1);
-    QVERIFY(!popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
     QVERIFY(!overlay->isVisible());
     QVERIFY(!button->isPressed());
     QCOMPARE(overlayPressedSignal.count(), overlayPressCount);
@@ -334,6 +375,7 @@ void tst_QQuickPopup::zOrder()
 {
     QFETCH(QString, source);
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -356,12 +398,12 @@ void tst_QQuickPopup::zOrder()
     QVERIFY(popup->isVisible());
 
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
-    QVERIFY(!popup2->isVisible());
+    QTRY_VERIFY(!popup2->isVisible());
     QVERIFY(popup->isVisible());
 
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QVERIFY(!popup2->isVisible());
-    QVERIFY(!popup->isVisible());
+    QTRY_VERIFY(!popup->isVisible());
 }
 
 void tst_QQuickPopup::windowChange()
@@ -387,6 +429,28 @@ void tst_QQuickPopup::windowChange()
     popup.setParentItem(window.contentItem());
     QCOMPARE(popup.window(), &window);
     QCOMPARE(spy.count(), 3);
+
+    popup.resetParentItem();
+    QVERIFY(!popup.window());
+    QCOMPARE(spy.count(), 4);
+
+    popup.setParent(&window);
+    popup.resetParentItem();
+    QCOMPARE(popup.window(), &window);
+    QCOMPARE(spy.count(), 5);
+
+    popup.setParent(this);
+    popup.resetParentItem();
+    QVERIFY(!popup.window());
+    QCOMPARE(spy.count(), 6);
+
+    item.setParentItem(window.contentItem());
+    popup.setParent(&item);
+    popup.resetParentItem();
+    QCOMPARE(popup.window(), &window);
+    QCOMPARE(spy.count(), 7);
+
+    popup.setParent(nullptr);
 }
 
 Q_DECLARE_METATYPE(QQuickPopup::ClosePolicy)
@@ -421,6 +485,7 @@ void tst_QQuickPopup::closePolicy()
     QFETCH(QQuickPopup::ClosePolicy, closePolicy);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -439,49 +504,55 @@ void tst_QQuickPopup::closePolicy()
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     // press outside popup and its parent
-    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1), 50);
     if (closePolicy.testFlag(QQuickPopup::CloseOnPressOutside) || closePolicy.testFlag(QQuickPopup::CloseOnPressOutsideParent))
-        QVERIFY(!popup->isVisible());
+        QTRY_VERIFY(!popup->isVisible());
     else
         QVERIFY(popup->isVisible());
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     // release outside popup and its parent
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     if (closePolicy.testFlag(QQuickPopup::CloseOnReleaseOutside))
-        QVERIFY(!popup->isVisible());
+        QTRY_VERIFY(!popup->isVisible());
     else
         QVERIFY(popup->isVisible());
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     // press outside popup but inside its parent
-    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x(), button->y()));
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x() + 1, button->y() + 1));
     if (closePolicy.testFlag(QQuickPopup::CloseOnPressOutside) && !closePolicy.testFlag(QQuickPopup::CloseOnPressOutsideParent))
-        QVERIFY(!popup->isVisible());
+        QTRY_VERIFY(!popup->isVisible());
     else
         QVERIFY(popup->isVisible());
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     // release outside popup but inside its parent
-    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x(), button->y()));
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x() + 1, button->y() + 1));
     if (closePolicy.testFlag(QQuickPopup::CloseOnReleaseOutside) && !closePolicy.testFlag(QQuickPopup::CloseOnReleaseOutsideParent))
-        QVERIFY(!popup->isVisible());
+        QTRY_VERIFY(!popup->isVisible());
     else
         QVERIFY(popup->isVisible());
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     // press inside and release outside
-    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x() + popup->x(), button->y() + popup->y()));
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(button->x() + popup->x() + 1,
+                                                                     button->y() + popup->y() + 1));
     QVERIFY(popup->isVisible());
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(1, 1));
     QVERIFY(popup->isVisible());
@@ -489,7 +560,7 @@ void tst_QQuickPopup::closePolicy()
     // escape
     QTest::keyClick(window, Qt::Key_Escape);
     if (closePolicy.testFlag(QQuickPopup::CloseOnEscape))
-        QVERIFY(!popup->isVisible());
+        QTRY_VERIFY(!popup->isVisible());
     else
         QVERIFY(popup->isVisible());
 }
@@ -499,6 +570,7 @@ void tst_QQuickPopup::activeFocusOnClose1()
     // Test that a popup that never sets focus: true (e.g. ToolTip) doesn't affect
     // the active focus item when it closes.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose1.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -512,24 +584,27 @@ void tst_QQuickPopup::activeFocusOnClose1()
 
     focusedPopup->open();
     QVERIFY(focusedPopup->isVisible());
+    QTRY_VERIFY(focusedPopup->isOpened());
     QVERIFY(focusedPopup->hasActiveFocus());
 
     nonFocusedPopup->open();
     QVERIFY(nonFocusedPopup->isVisible());
+    QTRY_VERIFY(nonFocusedPopup->isOpened());
     QVERIFY(focusedPopup->hasActiveFocus());
 
     nonFocusedPopup->close();
-    QVERIFY(!nonFocusedPopup->isVisible());
+    QTRY_VERIFY(!nonFocusedPopup->isVisible());
     QVERIFY(focusedPopup->hasActiveFocus());
 
     // QTBUG-66113: force active focus on a popup that did not request focus
     nonFocusedPopup->open();
     nonFocusedPopup->forceActiveFocus();
     QVERIFY(nonFocusedPopup->isVisible());
+    QTRY_VERIFY(nonFocusedPopup->isOpened());
     QVERIFY(nonFocusedPopup->hasActiveFocus());
 
     nonFocusedPopup->close();
-    QVERIFY(!nonFocusedPopup->isVisible());
+    QTRY_VERIFY(!nonFocusedPopup->isVisible());
     QVERIFY(focusedPopup->hasActiveFocus());
 }
 
@@ -539,6 +614,7 @@ void tst_QQuickPopup::activeFocusOnClose2()
     // calling forceActiveFocus() on another item) before it closes doesn't
     // affect the active focus item when it closes.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose2.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -555,16 +631,18 @@ void tst_QQuickPopup::activeFocusOnClose2()
 
     popup1->open();
     QVERIFY(popup1->isVisible());
+    QTRY_VERIFY(popup1->isOpened());
     QVERIFY(popup1->hasActiveFocus());
 
     popup2->open();
     QVERIFY(popup2->isVisible());
+    QTRY_VERIFY(popup2->isOpened());
     QVERIFY(popup2->hasActiveFocus());
 
     // Causes popup1.contentItem.forceActiveFocus() to be called, then closes popup2.
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier,
         closePopup2Button->mapToScene(QPointF(closePopup2Button->width() / 2, closePopup2Button->height() / 2)).toPoint());
-    QVERIFY(!popup2->isVisible());
+    QTRY_VERIFY(!popup2->isVisible());
     QVERIFY(popup1->hasActiveFocus());
 }
 
@@ -573,6 +651,7 @@ void tst_QQuickPopup::activeFocusOnClose3()
     // Test that a closing popup that had focus doesn't steal focus from
     // another popup that the focus was transferred to.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose3.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -600,6 +679,55 @@ void tst_QQuickPopup::activeFocusOnClose3()
     QTRY_VERIFY(popup2->hasActiveFocus());
 }
 
+void tst_QQuickPopup::activeFocusOnClosingSeveralPopups()
+{
+    // Test that active focus isn't lost when multiple popup closing simultaneously
+    QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClosingSeveralPopups.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
+    QQuickApplicationWindow *window = helper.appWindow;
+    window->show();
+    window->requestActivate();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickItem *button = window->property("button").value<QQuickItem *>();
+    QVERIFY(button);
+
+    QQuickPopup *popup1 = window->property("popup1").value<QQuickPopup *>();
+    QVERIFY(popup1);
+
+    QQuickPopup *popup2 = window->property("popup2").value<QQuickPopup *>();
+    QVERIFY(popup2);
+
+    QCOMPARE(button->hasActiveFocus(), true);
+    popup1->open();
+    QTRY_VERIFY(popup1->isOpened());
+    QVERIFY(popup1->hasActiveFocus());
+    popup2->open();
+    QTRY_VERIFY(popup2->isOpened());
+    QVERIFY(popup2->hasActiveFocus());
+    QTRY_COMPARE(button->hasActiveFocus(), false);
+    // close the unfocused popup first
+    popup1->close();
+    popup2->close();
+    QTRY_VERIFY(!popup1->isVisible());
+    QTRY_VERIFY(!popup2->isVisible());
+    QTRY_COMPARE(button->hasActiveFocus(), true);
+
+    popup1->open();
+    QTRY_VERIFY(popup1->isOpened());
+    QVERIFY(popup1->hasActiveFocus());
+    popup2->open();
+    QTRY_VERIFY(popup2->isOpened());
+    QVERIFY(popup2->hasActiveFocus());
+    QTRY_COMPARE(button->hasActiveFocus(), false);
+    // close the focused popup first
+    popup2->close();
+    popup1->close();
+    QTRY_VERIFY(!popup1->isVisible());
+    QTRY_VERIFY(!popup2->isVisible());
+    QTRY_COMPARE(button->hasActiveFocus(), true);
+}
+
 void tst_QQuickPopup::hover_data()
 {
     QTest::addColumn<QString>("source");
@@ -617,6 +745,7 @@ void tst_QQuickPopup::hover()
     QFETCH(bool, modal);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     window->requestActivate();
@@ -650,7 +779,7 @@ void tst_QQuickPopup::hover()
     QVERIFY(!childButton->isHovered());
 
     // hover the child button in a popup
-    QTest::mouseMove(window, QPoint(2, 2));
+    QTest::mouseMove(window, QPoint(popup->x() + popup->width() / 2, popup->y() + popup->height() / 2));
     QVERIFY(!parentButton->isHovered());
     QVERIFY(childButton->isHovered());
 
@@ -678,7 +807,9 @@ void tst_QQuickPopup::wheel_data()
 static bool sendWheelEvent(QQuickItem *item, const QPoint &localPos, int degrees)
 {
     QQuickWindow *window = item->window();
-    QWheelEvent wheelEvent(localPos, item->window()->mapToGlobal(localPos), QPoint(0, 0), QPoint(0, 8 * degrees), 0, Qt::Vertical, Qt::NoButton, 0);
+    QWheelEvent wheelEvent(localPos, item->window()->mapToGlobal(localPos),
+                           QPoint(0, 0), QPoint(0, 8 * degrees),
+                           0, Qt::Vertical, Qt::NoButton, {});
     QSpontaneKeyEvent::setSpontaneous(&wheelEvent);
     return qGuiApp->notify(window, &wheelEvent);
 }
@@ -689,6 +820,7 @@ void tst_QQuickPopup::wheel()
     QFETCH(bool, modal);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -759,6 +891,7 @@ void tst_QQuickPopup::parentDestroyed()
 void tst_QQuickPopup::nested()
 {
     QQuickApplicationHelper helper(this, QStringLiteral("nested.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -786,6 +919,7 @@ void tst_QQuickPopup::nested()
 void tst_QQuickPopup::grabber()
 {
     QQuickApplicationHelper helper(this, QStringLiteral("grabber.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -800,35 +934,35 @@ void tst_QQuickPopup::grabber()
     QVERIFY(combo);
 
     menu->open();
-    QCOMPARE(menu->isVisible(), true);
+    QTRY_COMPARE(menu->isOpened(), true);
     QCOMPARE(popup->isVisible(), false);
     QCOMPARE(combo->isVisible(), false);
 
     // click a menu item to open the popup
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(menu->width() / 2, menu->height() / 2));
-    QCOMPARE(menu->isVisible(), false);
-    QCOMPARE(popup->isVisible(), true);
+    QTRY_COMPARE(menu->isVisible(), false);
+    QTRY_COMPARE(popup->isOpened(), true);
     QCOMPARE(combo->isVisible(), false);
 
     combo->open();
     QCOMPARE(menu->isVisible(), false);
     QCOMPARE(popup->isVisible(), true);
-    QCOMPARE(combo->isVisible(), true);
+    QTRY_COMPARE(combo->isOpened(), true);
 
     // click outside to close both the combo popup and the parent popup
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - 1, window->height() - 1));
     QCOMPARE(menu->isVisible(), false);
-    QCOMPARE(popup->isVisible(), false);
-    QCOMPARE(combo->isVisible(), false);
+    QTRY_COMPARE(popup->isVisible(), false);
+    QTRY_COMPARE(combo->isVisible(), false);
 
     menu->open();
-    QCOMPARE(menu->isVisible(), true);
+    QTRY_COMPARE(menu->isOpened(), true);
     QCOMPARE(popup->isVisible(), false);
     QCOMPARE(combo->isVisible(), false);
 
     // click outside the menu to close it (QTBUG-56697)
     QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - 1, window->height() - 1));
-    QCOMPARE(menu->isVisible(), false);
+    QTRY_COMPARE(menu->isVisible(), false);
     QCOMPARE(popup->isVisible(), false);
     QCOMPARE(combo->isVisible(), false);
 }
@@ -838,7 +972,10 @@ void tst_QQuickPopup::cursorShape()
     // Ensure that the mouse cursor has the correct shape when over a popup
     // which is itself over an item with a different shape.
     QQuickApplicationHelper helper(this, QStringLiteral("cursor.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
+    centerOnScreen(window);
+    moveMouseAway(window);
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
 
@@ -847,17 +984,18 @@ void tst_QQuickPopup::cursorShape()
 
     popup->open();
     QVERIFY(popup->isVisible());
+    QTRY_VERIFY(popup->isOpened());
 
     QQuickItem *textField = helper.appWindow->property("textField").value<QQuickItem*>();
     QVERIFY(textField);
 
     // Move the mouse over the text field.
-    const QPoint textFieldPos(popup->x() - 10, popup->y() + popup->height() / 2);
+    const QPoint textFieldPos(popup->x() - 10, textField->height() / 2);
     QTest::mouseMove(window, textFieldPos);
     QCOMPARE(window->cursor().shape(), textField->cursor().shape());
 
     // Move the mouse over the popup where it overlaps with the text field.
-    const QPoint textFieldOverlapPos(popup->x() + 10, popup->y() + popup->height() / 2);
+    const QPoint textFieldOverlapPos(popup->x() + 10, textField->height() / 2);
     QTest::mouseMove(window, textFieldOverlapPos);
     QCOMPARE(window->cursor().shape(), popup->popupItem()->cursor().shape());
 
@@ -892,6 +1030,7 @@ void tst_QQuickPopup::closeOnEscapeWithNestedPopups()
     // Tests the scenario in the Gallery example, where there are nested popups that should
     // close in the correct order when the Escape key is pressed.
     QQuickApplicationHelper helper(this, QStringLiteral("closeOnEscapeWithNestedPopups.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -952,6 +1091,43 @@ void tst_QQuickPopup::closeOnEscapeWithNestedPopups()
     QCOMPARE(stackView->depth(), 1);
 }
 
+void tst_QQuickPopup::closeOnEscapeWithVisiblePopup()
+{
+    QQuickApplicationHelper helper(this, QStringLiteral("closeOnEscapeWithVisiblePopup.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickPopup *popup = window->findChild<QQuickPopup *>("popup");
+    QVERIFY(popup);
+    QTRY_VERIFY(popup->isOpened());
+
+    QTRY_VERIFY(window->activeFocusItem());
+    QTest::keyClick(window, Qt::Key_Escape);
+    QTRY_VERIFY(!popup->isVisible());
+}
+
+void tst_QQuickPopup::enabled()
+{
+    QQuickPopup popup;
+    QVERIFY(popup.isEnabled());
+    QVERIFY(popup.popupItem()->isEnabled());
+
+    QSignalSpy enabledSpy(&popup, &QQuickPopup::enabledChanged);
+    QVERIFY(enabledSpy.isValid());
+
+    popup.setEnabled(false);
+    QVERIFY(!popup.isEnabled());
+    QVERIFY(!popup.popupItem()->isEnabled());
+    QCOMPARE(enabledSpy.count(), 1);
+
+    popup.popupItem()->setEnabled(true);
+    QVERIFY(popup.isEnabled());
+    QVERIFY(popup.popupItem()->isEnabled());
+    QCOMPARE(enabledSpy.count(), 2);
+}
+
 void tst_QQuickPopup::orientation_data()
 {
     QTest::addColumn<Qt::ScreenOrientation>("orientation");
@@ -969,6 +1145,7 @@ void tst_QQuickPopup::orientation()
     QFETCH(QPointF, position);
 
     QQuickApplicationHelper helper(this, "orientation.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->reportContentOrientationChange(orientation);
@@ -982,6 +1159,306 @@ void tst_QQuickPopup::orientation()
     QCOMPARE(popup->popupItem()->position(), position);
 }
 
-QTEST_MAIN(tst_QQuickPopup)
+void tst_QQuickPopup::qquickview()
+{
+    QQuickView view;
+    view.setObjectName("QQuickView");
+    view.resize(400, 400);
+    view.setSource(testFileUrl("dialog.qml"));
+    QVERIFY(view.status() != QQuickView::Error);
+    view.contentItem()->setObjectName("QQuickViewContentItem");
+    view.show();
+
+    QQuickDialog *dialog = view.rootObject()->property("dialog").value<QQuickDialog*>();
+    QVERIFY(dialog);
+    QTRY_COMPARE(dialog->property("opened").toBool(), true);
+
+    dialog->close();
+    QTRY_COMPARE(dialog->property("visible").toBool(), false);
+
+    // QTBUG-72746: shouldn't crash on application exit after closing a Dialog when using QQuickView.
+}
+
+// TODO: also test it out without setting enabled directly on menu, but on a parent
+
+// QTBUG-73447
+void tst_QQuickPopup::disabledPalette()
+{
+    QQuickApplicationHelper helper(this, "disabledPalette.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
+    QVERIFY(popup);
+
+    QSignalSpy popupEnabledSpy(popup, SIGNAL(enabledChanged()));
+    QVERIFY(popupEnabledSpy.isValid());
+    QSignalSpy popupPaletteSpy(popup, SIGNAL(paletteChanged()));
+    QVERIFY(popupPaletteSpy.isValid());
+
+    QSignalSpy popupItemEnabledSpy(popup->popupItem(), SIGNAL(enabledChanged()));
+    QVERIFY(popupItemEnabledSpy.isValid());
+    QSignalSpy popupItemPaletteSpy(popup->popupItem(), SIGNAL(paletteChanged()));
+    QVERIFY(popupItemPaletteSpy.isValid());
+
+    QPalette palette = popup->palette();
+    palette.setColor(QPalette::Active, QPalette::Base, Qt::green);
+    palette.setColor(QPalette::Disabled, QPalette::Base, Qt::red);
+    popup->setPalette(palette);
+    QCOMPARE(popupPaletteSpy.count(), 1);
+    QCOMPARE(popupItemPaletteSpy.count(), 1);
+    QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::green);
+
+    popup->setEnabled(false);
+    QCOMPARE(popupEnabledSpy.count(), 1);
+    QCOMPARE(popupItemEnabledSpy.count(), 1);
+    QCOMPARE(popupPaletteSpy.count(), 2);
+    QCOMPARE(popupItemPaletteSpy.count(), 2);
+    QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::red);
+}
+
+void tst_QQuickPopup::disabledParentPalette()
+{
+    QQuickApplicationHelper helper(this, "disabledPalette.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
+    QVERIFY(popup);
+
+    QSignalSpy popupEnabledSpy(popup, SIGNAL(enabledChanged()));
+    QVERIFY(popupEnabledSpy.isValid());
+    QSignalSpy popupPaletteSpy(popup, SIGNAL(paletteChanged()));
+    QVERIFY(popupPaletteSpy.isValid());
+
+    QSignalSpy popupItemEnabledSpy(popup->popupItem(), SIGNAL(enabledChanged()));
+    QVERIFY(popupItemEnabledSpy.isValid());
+    QSignalSpy popupItemPaletteSpy(popup->popupItem(), SIGNAL(paletteChanged()));
+    QVERIFY(popupItemPaletteSpy.isValid());
+
+    QPalette palette = popup->palette();
+    palette.setColor(QPalette::Active, QPalette::Base, Qt::green);
+    palette.setColor(QPalette::Disabled, QPalette::Base, Qt::red);
+    popup->setPalette(palette);
+    QCOMPARE(popupPaletteSpy.count(), 1);
+    QCOMPARE(popupItemPaletteSpy.count(), 1);
+    QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::green);
+
+    // Disable the overlay (which is QQuickPopupItem's parent) to ensure that
+    // the palette is changed when the popup is indirectly disabled.
+    popup->open();
+    QTRY_VERIFY(popup->isOpened());
+    QVERIFY(QMetaObject::invokeMethod(window, "disableOverlay"));
+    QVERIFY(!popup->isEnabled());
+    QVERIFY(!popup->popupItem()->isEnabled());
+    QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::red);
+    QCOMPARE(popupEnabledSpy.count(), 1);
+    QCOMPARE(popupItemEnabledSpy.count(), 1);
+    QCOMPARE(popupPaletteSpy.count(), 2);
+    QCOMPARE(popupItemPaletteSpy.count(), 2);
+
+    popup->close();
+    QTRY_VERIFY(!popup->isVisible());
+}
+
+void tst_QQuickPopup::countChanged()
+{
+    QQuickApplicationHelper helper(this, "countChanged.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickComboBox *comboBox = window->property("comboBox").value<QQuickComboBox*>();
+    QVERIFY(comboBox);
+    QCOMPARE(window->property("count").toInt(), 1);
+
+    QVERIFY(window->setProperty("isModel1", false));
+    QTRY_COMPARE(window->property("count").toInt(), 2);
+}
+
+// QTBUG-73243
+void tst_QQuickPopup::toolTipCrashOnClose()
+{
+    QQuickApplicationHelper helper(this, "toolTipCrashOnClose.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    // The warning only occurs with debug builds for some reason.
+    // In any case, the warning is irrelevant, but using ShaderEffectSource is important, so we ignore it.
+#ifdef QT_DEBUG
+    QTest::ignoreMessage(QtWarningMsg, "ShaderEffectSource: 'recursive' must be set to true when rendering recursively.");
+#endif
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QTest::mouseMove(window, QPoint(window->width() / 2, window->height() / 2));
+    QTRY_VERIFY(window->property("toolTipOpened").toBool());
+
+    QVERIFY(window->close());
+    // Shouldn't crash.
+}
+
+void tst_QQuickPopup::setOverlayParentToNull()
+{
+    QQuickApplicationHelper helper(this, "toolTipCrashOnClose.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    centerOnScreen(window);
+    moveMouseAway(window);
+    window->show();
+#ifdef QT_DEBUG
+    QTest::ignoreMessage(QtWarningMsg, "ShaderEffectSource: 'recursive' must be set to true when rendering recursively.");
+#endif
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QVERIFY(QMetaObject::invokeMethod(window, "nullifyOverlayParent"));
+
+    QTest::mouseMove(window, QPoint(window->width() / 2, window->height() / 2));
+    QTRY_VERIFY(window->property("toolTipOpened").toBool());
+
+    QVERIFY(window->close());
+    // While nullifying the overlay parent doesn't make much sense, it shouldn't crash.
+}
+
+void tst_QQuickPopup::tabFence()
+{
+    if (QGuiApplication::styleHints()->tabFocusBehavior() != Qt::TabFocusAllControls)
+        QSKIP("This platform only allows tab focus for text controls");
+
+    QQuickApplicationHelper helper(this, "tabFence.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickPopup *popup = window->property("dialog").value<QQuickPopup*>();
+    QVERIFY(popup);
+    popup->open();
+    popup->setModal(true);
+
+    QQuickButton *outsideButton1 = window->property("outsideButton1").value<QQuickButton*>();
+    QVERIFY(outsideButton1);
+    QQuickButton *outsideButton2 = window->property("outsideButton2").value<QQuickButton*>();
+    QVERIFY(outsideButton2);
+    QQuickButton *dialogButton1 = window->property("dialogButton1").value<QQuickButton*>();
+    QVERIFY(dialogButton1);
+    QQuickButton *dialogButton2 = window->property("dialogButton2").value<QQuickButton*>();
+    QVERIFY(dialogButton2);
+
+    // When modal, focus loops between the two external buttons
+    outsideButton1->forceActiveFocus();
+    QVERIFY(outsideButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton1->hasActiveFocus());
+
+    // Same thing for dialog's buttons
+    dialogButton1->forceActiveFocus();
+    QVERIFY(dialogButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton1->hasActiveFocus());
+
+    popup->setModal(false);
+
+    // When not modal, focus goes in and out of the dialog
+    outsideButton1->forceActiveFocus();
+    QVERIFY(outsideButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton1->hasActiveFocus());
+}
+
+void tst_QQuickPopup::invisibleToolTipOpen()
+{
+    QQuickApplicationHelper helper(this, "invisibleToolTipOpen.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    centerOnScreen(window);
+    moveMouseAway(window);
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickItem *mouseArea = qvariant_cast<QQuickItem *>(window->property("mouseArea"));
+    QVERIFY(mouseArea);
+    QObject *loader = qvariant_cast<QObject *>(window->property("loader"));
+    QVERIFY(loader);
+
+    QTest::mouseMove(window, QPoint(mouseArea->width() / 2, mouseArea->height() / 2));
+    QTRY_VERIFY(mouseArea->property("isToolTipVisible").toBool());
+
+    QSignalSpy componentLoadedSpy(loader, SIGNAL(loaded()));
+    QVERIFY(componentLoadedSpy.isValid());
+
+    loader->setProperty("active", true);
+    QTRY_COMPARE(componentLoadedSpy.count(), 1);
+
+    QTRY_VERIFY(mouseArea->property("isToolTipVisible").toBool());
+}
+
+void tst_QQuickPopup::centerInOverlayWithinStackViewItem()
+{
+    QQuickApplicationHelper helper(this, "centerInOverlayWithinStackViewItem.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
+    QVERIFY(popup);
+    QTRY_COMPARE(popup->isVisible(), true);
+
+    // Shouldn't crash on exit.
+}
+
+void tst_QQuickPopup::destroyDuringExitTransition()
+{
+    QQuickApplicationHelper helper(this, "destroyDuringExitTransition.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QPointer<QQuickPopup> dialog2 = window->property("dialog2").value<QQuickPopup*>();
+    QVERIFY(dialog2);
+    QTRY_COMPARE(dialog2->isVisible(), true);
+
+    // Close the second dialog, destroying it before its exit transition can finish.
+    QTest::keyClick(window, Qt::Key_Escape);
+    QTRY_VERIFY(!dialog2);
+
+    // Events should go through to the dialog underneath.
+    QQuickPopup *dialog1 = window->property("dialog1").value<QQuickPopup*>();
+    QVERIFY(dialog1);
+    QQuickButton *button = dialog1->property("button").value<QQuickButton*>();
+    QVERIFY(button);
+    const auto buttonClickPos = button->mapToScene(QPointF(button->width() / 2, button->height() / 2)).toPoint();
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, buttonClickPos);
+    QVERIFY(button->isDown());
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, buttonClickPos);
+    QVERIFY(!button->isDown());
+}
+
+QTEST_QUICKCONTROLS_MAIN(tst_QQuickPopup)
 
 #include "tst_qquickpopup.moc"
